import uuid
import sass
import os

from django.conf import settings
from django.core.files.base import ContentFile
from django.db import models
from django.db.models.signals import post_save
from django.dispatch import receiver
from django.utils.translation import ugettext as _
from django.urls import reverse
from django.contrib.sites.models import Site
from django.utils.text import slugify

from autoslug import AutoSlugField
from tempfile import gettempdir

STYLIST_CSS_MEDIA_PATH = getattr(settings, "STYLIST_CSS_MEDIA_PATH", 'site/{domain}/style/{filename}')
STYLIST_DEFAULT_CSS = getattr(settings, "STYLIST_DEFAULT_CSS")
STYLIST_SCSS_TEMPLATE = getattr(settings, "STYLIST_SCSS_TEMPLATE")

def css_file_path(instance, filename):
    """
    Method sets up the path to save the CSS file to.
    """
    # return 'site/{domain}/style/{filename}'.format(context)

    context = { 'domain': instance.site.domain,      # Done this way so we can expose more varibles
                'filename': filename,
                'site_name': slugify(instance.site.name),
                'uuid': instance.uuid}

    return STYLIST_CSS_MEDIA_PATH.format(**context)

def default_attrs():
    attrs = {}
    for key in settings.STYLE_SCHEMA:
        attrs[key] = settings.STYLE_SCHEMA[key]["default"]
    return attrs


class Style(models.Model):
    """
    The model that holds all the variables for 
    """
    name = models.CharField(_("Name"), max_length=50)
    uuid = models.UUIDField(default=uuid.uuid4, editable=False, unique=True)
    slug = AutoSlugField(verbose_name=_("Slug"), populate_from="name", unique_with=('site__id',), always_update=True)
    site = models.ForeignKey(Site, on_delete=models.CASCADE, blank=True, null=True, related_name="site_style")        # Overwriting this field from Base Class to change the related_name
    attrs = models.JSONField(blank=True, default=default_attrs)
    enabled = models.BooleanField(_("Enabled"))
    css_file = models.FileField(_("CSS File"), upload_to=css_file_path, max_length=100, blank=True, null=True)                                 # Location where the Compiled CSS file is placed.  Used so a completely custom file can be uploaded.

    class Meta:
        verbose_name = _( "Site Style")
        verbose_name_plural = _( "Site Styles")

    def __str__(self):
        return "{}: {}".format(self.site.name, self.name)

    def get_absolute_url(self):
        return reverse( "stylist:stylist-edit", kwargs={"uuid": self.uuid})
    
    def get_style_css(self):
        if not self.css_file:
            return STYLIST_DEFAULT_CSS
        
        return self.css_file.url

    def compile_attrs(self):
        with open(gettempdir() + "/custom_vars.scss", "w+") as custom_vars:
            string = ""
            google_fonts = "@import url('https://fonts.googleapis.com/css2?family="
            num_fonts = 0
            for key in self.attrs:
                string += "$" + key + ": " + self.attrs[key] + ";\n"
                if settings.STYLE_SCHEMA[key]["type"] == "font":
                    if num_fonts > 0:
                        google_fonts += "&family="
                    google_fonts += self.attrs[key].replace(" ", "+")
                    google_fonts += ":wght@100;200;300;400;500;600;700;800;900"
                    num_fonts += 1
            if num_fonts > 0:
                google_fonts += "&display=swap');\n"
                string = google_fonts + string
            custom_vars.write(string)
            custom_vars.seek(0)
            
            self.css_file.save(self.name + ".css", ContentFile(sass.compile(filename=STYLIST_SCSS_TEMPLATE, include_paths=[gettempdir()])))
            os.remove(custom_vars.name)


@receiver(post_save, sender=Style)
def only_one_active_style(sender, instance, **kwargs):
    if instance.enabled:
        Style.objects.filter(site=instance.site).exclude(pk=instance.pk).update(enabled=False)



# Generated by Django 3.1.2 on 2020-10-09 17:11

import autoslug.fields
from django.db import migrations, models
import django.db.models.deletion
import stylist.models
import uuid


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('sites', '0002_alter_domain_unique'),
    ]

    operations = [
        migrations.CreateModel(
            name='Style',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=50, verbose_name='Name')),
                ('uuid', models.UUIDField(default=uuid.uuid4, editable=False, unique=True)),
                ('slug', autoslug.fields.AutoSlugField(always_update=True, editable=False, populate_from='name', unique_with=('site__id',), verbose_name='Slug')),
                ('attrs', models.JSONField(blank=True, default=dict)),
                ('enabled', models.BooleanField(verbose_name='Enabled')),
                ('css_file', models.FileField(upload_to=stylist.models.css_file_path, verbose_name='CSS File')),
                ('site', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='site_style', to='sites.site')),
            ],
            options={
                'verbose_name': 'Site Style',
                'verbose_name_plural': 'Site Styles',
            },
        ),
    ]



from rest_framework.generics import CreateAPIView
from django.contrib.sites.models import Site
from django.shortcuts import redirect

from stylist.models import Style 
from stylist.api.serializers import StyleSerializer


class StyleCreateAPIView(CreateAPIView):
     queryset = Style.objects.all()
     serializer_class = StyleSerializer

     def perform_create(self, serializer):
          site = Site.objects.get_current()
          instance = serializer.save(site=site)
          instance.compile_attrs()

          if not Style.objects.filter(site=site, enabled=True):
               instance.enabled = True
               instance.save()

     def create(self, request, *args, **kwargs):
          response = super().create(request, *args, **kwargs)
          return redirect('stylist:stylist-index')



from django.urls import path

from stylist import views

app_name = "stylist"

urlpatterns = [
    path("", views.StylistIndexView.as_view(), name="stylist-index"),
    path("style/<uuid:uuid>/edit/", views.StylistUpdateView.as_view(), name="stylist-edit"),
    path("style/<uuid:uuid>/preview/", views.StylistPreviewView.as_view(), name="stylist-preview"),
    path("style/<uuid:uuid>/delete/", views.StylistDeleteView.as_view(), name="stylist-delete"),
    path("style/active/", views.StylistActiveView.as_view(), name="stylist-make-active"),
    path("style/preview/end/", views.end_preview, name="stylist-end-preview"),
]



import os

from io import StringIO

from django.conf import settings
from django.contrib.sites.models import Site
from django.core.files.storage import default_storage
from django.core.management import call_command
from django.test import TestCase, Client
from django.urls import reverse

from stylist.models import Style

# Create your tests here.
class ModuleTests(TestCase):
    '''
    Tests to make sure basic elements are not missing from the package
    '''
    def test_for_missing_migrations(self):
        output = StringIO()

        try:
            call_command('makemigrations', interactive=False, dry_run=True, check=True, stdout=output)

        except SystemExit as e:
            # The exit code will be 1 when there are no missing migrations
            try:
                assert e == '1'
            except:
                self.fail("\n\nHey, There are missing migrations!\n\n %s" % output.getvalue())


class StylistClientTests(TestCase):
    '''
    Client tests for Stylist and the Style model
    '''
    def setUp(self):
        self.client = Client()
    
    def test_style_creation_api(self):
        try:
            self.assertEquals(Style.objects.all().count(), 0)
            url = reverse("api-create-style")
            response = self.client.post(url, {"name": "Test", "enabled": "False"}, follow=True)
            self.assertEquals(Style.objects.all().count(), 1)
            
            style = Style.objects.get(pk=1)
            # since it's the only style, it should be enabled
            self.assertEquals(style.enabled, True)

            site = Site.objects.get_current()
            self.assertEquals(style.site, site)
            css_path = settings.MEDIA_ROOT + "/site/" + site.domain + "/style/Test.css"
            self.assertEquals(style.css_file.path, css_path)
            # remove the new file, leaving any preexisting files
            if(os.path.exists(css_path)):
                os.remove(css_path)
        except:
            print("")
            print(response.status_code)
            print(response.content.decode('utf-8'))
            raise

    def test_style_update(self):
        try:
            site = Site.objects.get_current()
            style = Style.objects.create(name="Test", enabled=False, site=site)
            url = reverse("stylist:stylist-edit", kwargs={"uuid": style.uuid})
            
            data = style.attrs
            data["name"] = "Updated"
            data["primary"] = "#FF0000"
            response = self.client.post(url, data, follow=True)

            style.refresh_from_db()
            self.assertEquals(style.name, "Updated")
            self.assertEquals(style.attrs["primary"], "#FF0000")
            
            css_path = settings.MEDIA_ROOT + "/site/" + site.domain + "/style/Updated.css"
            # remove the new file, leaving any preexisting files
            if(os.path.exists(css_path)):
                os.remove(css_path)
            
        except:
            print("")
            print(response.status_code)
            print(response.content.decode('utf-8'))
            raise

    def test_style_delete(self):
        try:
            style = Style.objects.create(name="Test", enabled=True)
            self.assertEquals(Style.objects.all().count(), 1)
            
            url = reverse("stylist:stylist-delete", kwargs={"uuid": style.uuid})
            response = self.client.post(url, follow=True)

            self.assertEquals(Style.objects.all().count(), 0)
        except:
            print("")
            print(response.status_code)
            print(response.content.decode('utf-8'))
            raise

    def test_style_active(self):
        try:
            site = Site.objects.get_current()
            style = Style.objects.create(name="Test", enabled=False, site=site)
            url = reverse("stylist:stylist-make-active")

            response = self.client.post(url, {"active": 1}, follow=True)
            self.assertEquals(response.context["active_theme"], style)
        except:
            print("")
            print(response.status_code)
            print(response.content.decode('utf-8'))
            raise


    def test_style_preview(self):
        try:
            site = Site.objects.get_current()
            style = Style.objects.create(name="Test", enabled=False, site=site)

            url = reverse("stylist:stylist-preview", kwargs={"uuid": style.uuid})
            data = style.attrs
            data["name"] = style.name
            response = self.client.post(url, data, follow=True)
            self.assertEquals(response.context["preview_style"], settings.MEDIA_URL + "site/" + site.domain + "/style/Test_preview.css")

            # cleanup preview file
            path = self.client.session["preview_path"]
            default_storage.delete(path)
        except:
            print("")
            print(response.status_code)
            print(response.content.decode('utf-8'))
            raise


    def test_end_preview(self):
        try:
            site = Site.objects.get_current()
            style = Style.objects.create(name="Test", enabled=False, site=site)

            url = reverse("stylist:stylist-preview", kwargs={"uuid": style.uuid})
            data = style.attrs
            data["name"] = style.name
            response = self.client.post(url, data, follow=True)

            new_url = reverse("stylist:stylist-end-preview")
            new_response = self.client.get(new_url, HTTP_REFERER=reverse("stylist:stylist-index"), follow=True)
            self.assertEquals(new_response.context.get("preview_style"), None)
        except:
            print("")
            print(response.status_code)
            print(response.content.decode('utf-8'))
            raise



from django.contrib.sites.models import Site

from .models import Style

# ---------------
# CONTEXT PROCESSOR
# ---------------


def get_custom_styles(request):
    if getattr(request, "site", None):
        site = request.site
    else:
        site = Site.objects.get_current()
        
    try:
        style = Style.objects.filter(site=site).get(enabled=True)
        custom_style = style.css_file.url
    except:
        custom_style = None

    preview = request.session.get("preview_css")

    return {
        'custom_style': custom_style,
        'preview_style': preview,
        }




from django.contrib import admin

from .models import Style


def compile_styles(modeladmin, request, queryset):
    for q in queryset:
        q.compile_attrs()
compile_styles.short_description = "Compile selected styles"


class StyleAdmin(admin.ModelAdmin):
    list_display = ('id', 'name', 'slug', 'site', 'attrs', 'enabled')
    list_filter = ('site', 'enabled')
    search_fields = ('name', 'slug')
    actions = [compile_styles]


admin.site.register(Style, StyleAdmin)



import os
import sass

from django.conf import settings
from django.contrib.sites.models import Site
from django.core.files.base import ContentFile
from django.core.files.storage import default_storage
from django.shortcuts import reverse, redirect
from django.urls import reverse_lazy
from django.views.generic import FormView, ListView, UpdateView, DeleteView

from tempfile import gettempdir

from .forms import StyleForm, StyleEditForm, ActiveStyleForm
from .models import Style, css_file_path

class StylistIndexView(ListView):
    """
    List of styles available for editing
    """
    template_name = "stylist/index.html"
    model = Style

    def get_context_data(self, *args, **kwargs):
        context = super().get_context_data()
        context["form"] = StyleForm
        context["active_form"] = ActiveStyleForm
        site_id = settings.SITE_ID
        if hasattr(self.request, 'site'):
            site_id = self.request.site.id
        try:
            context["active_theme"] = Style.objects.filter(site=site_id).get(enabled=True)
        except:
            pass
        return context
    

# Stylist Preview Link - Autogenerated CSS from SASS/SCSS
class StylistPreviewView(FormView):
    form_class = StyleEditForm
    template_name = "stylist/style_form.html"
    success_url = reverse_lazy("stylist:stylist-index")

    def get_context_data(self, *args, **kwargs):
        context = super().get_context_data()
        context["object"] = Style.objects.get(uuid=self.kwargs["uuid"])
        return context

    def form_valid(self, form):
        with open(gettempdir() + "/custom_vars.scss", "w+") as custom_vars:
            string = ""
            google_fonts = "@import url('https://fonts.googleapis.com/css2?family="
            num_fonts = 0
            for key in form.cleaned_data:
                if key != "name":
                    string += "$" + key + ": " + form.cleaned_data[key] + ";\n"
                    if settings.STYLE_SCHEMA[key]["type"] == "font":
                        if num_fonts > 0:
                            google_fonts += "&family="
                        google_fonts += form.cleaned_data[key].replace(" ", "+")
                        google_fonts += ":wght@100;200;300;400;500;600;700;800;900"
                        num_fonts += 1
            if num_fonts > 0:
                google_fonts += "&display=swap');\n"
                string = google_fonts + string
            custom_vars.write(string)
            custom_vars.seek(0)

            instance = Style.objects.get(uuid=self.kwargs["uuid"])

            filename = settings.MEDIA_ROOT + "/" + css_file_path(instance, instance.name + "_preview.css")
            if default_storage.exists(filename):
                default_storage.delete(filename)
            preview = default_storage.save(filename, ContentFile(sass.compile(filename=settings.STYLIST_SCSS_TEMPLATE, include_paths=[gettempdir()])))
            
            self.request.session["preview_css"] = settings.MEDIA_URL + css_file_path(instance, instance.name + "_preview.css")
            self.request.session["preview_path"] = preview
            os.remove(custom_vars.name)
        return redirect(self.get_success_url())


# Stylist Edit Page
class StylistUpdateView(UpdateView):
    model = Style
    form_class = StyleEditForm

    def get_object(self, *args, **kwargs):
        return Style.objects.get(uuid=self.kwargs.get("uuid"))


# Change which Style is currently active for the site
class StylistActiveView(FormView):
    form_class = ActiveStyleForm
    template_name = "stylist/active_form.html"

    def form_valid(self, form):
        instance = form.cleaned_data["active"]
        instance.enabled = True
        instance.save()
        return redirect('stylist:stylist-index')

    def get_context_data(self, *args, **kwargs):
        context = super().get_context_data()
        site_id = settings.SITE_ID
        if hasattr(self.request, 'site'):
            site_id = self.request.site.id
        try:
            context["active_theme"] = Style.objects.filter(site=site_id).get(enabled=True)
        except:
            pass
        return context

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        kwargs['request'] = self.request
        return kwargs


# Stylist Detail Page


class StylistDeleteView(DeleteView):
    model = Style
    template_name = 'stylist/style_delete.html'

    def get_object(self, queryset=None):
        return Style.objects.get(uuid=self.kwargs.get("uuid"))

    def get_success_url(self):
        return reverse('stylist:stylist-index')


# Ends preview mode
def end_preview(request):
    preview = request.session.pop("preview_css", None)
    preview_path = request.session.pop("preview_path", None)
    if preview_path:
        default_storage.delete(preview_path)
    return redirect(request.META.get('HTTP_REFERER'))
